# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from earthkit.maps.styles import colors


def test_expand():
    assert colors.expand(["red"], [1, 2]) == ["red"]

    assert colors.expand("viridis", [1, 2])[0] == pytest.approx(
        [0.267, 0.005, 0.329, 1.000], 0.1
    )

    assert colors.expand("viridis", [1, 2])[1] == pytest.approx(
        [0.993, 0.906, 0.144, 1.000], 0.1
    )
