# Copyright 2023, European Centre for Medium Range Weather Forecasts.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cartopy.crs as ccrs

from earthkit.maps.domains import domain


def test_Domain_from_string():
    greenland = domain.Domain.from_string("greenland")
    assert greenland.title == "Greenland"
    assert isinstance(greenland.crs, ccrs.AlbersEqualArea)

    europe = domain.Domain.from_string("europe")
    assert europe.title == "Europe"
    assert isinstance(europe.crs, ccrs.LambertAzimuthalEqualArea)
